﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reactive.Concurrency;
using System.Reactive.PlatformServices;
using System.Reactive.Linq;
using System.Reactive.Subjects;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using System.Windows.Shapes;

using Microsoft.Research.DynamicDataDisplay;
using Microsoft.Research.DynamicDataDisplay.DataSources;

using ODataSampleClient.Utilities;

namespace ODataSampleClient
{
    /// <summary>
    /// Interaction logic for DataDisplay.xaml
    /// </summary>
    public partial class DataDisplay : Window
    {
        #region Private fields
        /// <summary>
        /// We can form connections to the OData service dynamically.
        /// </summary>
        private Uri dataLocation;

        /// <summary>
        /// The subscription in the OData service we are monitoring.
        /// </summary>
        private int subscriptionNo;

        /// <summary>
        /// When starting, collect the items currently monitored in this subscription.
        /// </summary>
        private List<Tuple<string, int>> itemsInSubscription;

        /// <summary>
        /// For each item monitored in this subscription, create a graphing sample buffer.
        /// </summary>
        private List<ItemSampleData> itemSamplesFromSubscription;

        /// <summary>
        /// The shared x-axis data buffer.
        /// </summary>
        readonly double[] animatedX = new double[1000];

        /// <summary>Programmatically created header</summary>
        Header chartHeader = new Header();

        /// <summary>Text contents of header</summary>
        TextBlock headerContents = new TextBlock();

        /// <summary>Timer to animate data</summary>
        readonly DispatcherTimer timer = new DispatcherTimer();
        System.Reactive.Concurrency.SynchronizationContextScheduler scheduler = null;
        #endregion

        #region Internal class ItemSampleData
        /// <summary>
        /// For a single item, capture the current sliding window
        /// of samples in a format acceptable to the DynamicDataDisplay library.
        /// </summary>
        public class ItemSampleData : IObserver<List<TimeSeriesDataFeed.Sample>>
        {
            #region Private fields
            /// <summary>
            /// The item id of the item being monitored.
            /// </summary>
            private int itemId;

            /// <summary>
            /// The buffer we have linked to DynamicDataDisplay for a pen.
            /// </summary>
            readonly double[] animatedY = new double[1000];

            /// <summary>
            /// A working buffer to accumulate samples from the service into.
            /// </summary>
            private List<double> samples = new List<double>();

            /// <summary>
            /// Each channel offsets the raw data by a fixed amount.
            /// </summary>
            private double offset = 0;

            /// <summary>
            /// The DynamicDataDisplay data source for this pen.
            /// </summary>
            private EnumerableDataSource<double> itemSamples = null;

            private IObservable<List<TimeSeriesDataFeed.Sample>> sampleFeed = null;
            System.Reactive.Concurrency.SynchronizationContextScheduler scheduler = null;
            #endregion

            #region Public properties
            /// <summary>
            /// The item id that this pen channel follows.
            /// </summary>
            public int ItemId { get { return itemId; } }

            /// <summary>
            /// Gets a reference to the DynamicDataDisplay data source.
            /// </summary>
            public EnumerableDataSource<double> ItemSamples { get { return itemSamples; } }

            /// <summary>
            /// Gets a reference to the raw samples buffer.
            /// </summary>
            public List<double> Samples { get { return samples; } }
            #endregion

            #region Constructor
            /// <summary>
            /// Construct the raw buffer and DynamicDataDisplay buffer
            /// for the pen channel for the item whose Id is given.
            /// </summary>
            /// <param name="id">The OData item id associated with this pen channel</param>
            public ItemSampleData(ODataSubscription subscription, SynchronizationContextScheduler s, int id, double offset)
            {
                // Capture the item id and offset for the pen channel.
                itemId = id;
                this.offset = offset;
                scheduler = s;

                // Initialize the internal raw and DyanamicDataDisplay buffers to zeros.
                for (int i = 0; i < animatedY.Length; i++)
                {
                    animatedY[i] = 0;
                    samples.Add(0);
                }

                // Make a DynamicDataDisplay data source and link it to the buffer.
                itemSamples = new EnumerableDataSource<double>(animatedY);
                itemSamples.SetYMapping(y => y);

                try
                {
                    sampleFeed = subscription.GetItemSampleFeed(id);
                    sampleFeed.Subscribe(this);
                }
                catch (ArgumentOutOfRangeException)
                {
                    sampleFeed = null;
                }
            }
            #endregion

            #region IObserver<TimeSeriesDataFeed.Sample> interface implementation
            // Summary:
            //     Notifies the observer that the provider has finished sending push-based notifications.
            public void OnCompleted()
            {

            }

            //
            // Summary:
            //     Notifies the observer that the provider has experienced an error condition.
            //
            // Parameters:
            //   error:
            //     An object that provides additional information about the error.
            public void OnError(Exception error)
            {

            }

            //
            // Summary:
            //     Provides the observer with new data.
            //
            // Parameters:
            //   value:
            //     The current notification information.
            public void OnNext(List<TimeSeriesDataFeed.Sample> values)
            {
                samples.AddRange((from value in values select (double)value.SampleValue + offset).ToList());
            }
            #endregion

            #region Private helper methods
            /// <summary>
            /// Dress the raw samples buffer, copy it to the DynamicDataDisplay buffer,
            /// and rais the data changed event.
            /// </summary>
            public void RaiseDataChange()
            {
                // If the raw buffer goes over the length of the DynamicDataDisplay buffer, trim it from the back.
                if (samples.Count > animatedY.Length)
                {
                    samples.RemoveRange(0, samples.Count - animatedY.Length);
                }

                // If we have the correct number of samples, copy the raw buffer to the DynamicDataDisplay buffer.
                if (samples.Count == animatedY.Length)
                {
                    samples.CopyTo(animatedY, 0);

                    // Here it is - signal that data is updated
                    itemSamples.RaiseDataChanged();
                }
            }
            #endregion
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes a new instance of the <see cref="DataDisplay"/> class.
        /// This represents a single OData subscription display, connecting
        /// all items to their own display channels on a DynamicDataDisplay pen chart.
        /// </summary>
        /// <param name="dataSourceLocation">The base URL to the ODATA source</param>
        /// <param name="subs">The subscription Id of the subscription to monitor</param>
        public DataDisplay(Uri dataSourceLocation, int subs)
        {
            // Capture the initializing parameters.
            dataLocation = dataSourceLocation;
            subscriptionNo = subs;
            itemsInSubscription = new List<Tuple<string, int>>();
            itemSamplesFromSubscription = new List<ItemSampleData>();

            // Let the base class initialize this Window.
            InitializeComponent();

            scheduler = new SynchronizationContextScheduler(System.Threading.SynchronizationContext.Current);

            // Set up the pen chart.
            headerContents.FontSize = 24;
            headerContents.Text = "Subscription " + subscriptionNo;
            headerContents.HorizontalAlignment = HorizontalAlignment.Center;
            chartHeader.Content = headerContents;
            plotter.Children.Add(chartHeader);

            // Set a handler into the Loaded event, continue setup after the Window is initialized.
            base.Loaded += Window_Loaded;

            // Handle closing notification to clean up.
            base.Closing += Window_Closing;
        }
        #endregion

        #region Plotting timer handler
        /// <summary>
        /// Periodically poll all items in the subscription, and
        /// update their pen chart displays.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void AnimatedPlot_Timer(object sender, EventArgs e)
        {
            // Sanity check that we have subscription information.
            if (itemsInSubscription == null || itemsInSubscription.Count == 0)
                return;

            // Notify each plotter to redraw.
            foreach (var item in itemSamplesFromSubscription)
            {
                item.RaiseDataChange();
            }
        }
        #endregion

        #region Window Loaded and Window Closing event handlers
        /// <summary>
        /// Loaded event handler.  When the WPF form is loaded,
        /// this event handler will discover the items in the
        /// subscription, and initialize the pen channels for each item.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            // Discover the items in this subscription.
            GetItems();

            // Set up the X axis labels.
            for (int i = 0; i < animatedX.Length; i++)
            {
                animatedX[i] = i;
            }

            // Make a DynamicDataDisplay shared source for the X axis.
            EnumerableDataSource<double> xSrc = new EnumerableDataSource<double>(animatedX);
            xSrc.SetXMapping(x => x);

            // Get a reference to the subscription feed for this subscription id.
            ODataSubscription.AddSubscription(subscriptionNo);
            ODataSubscription subscription = ODataSubscription.GetSubscription(subscriptionNo);

            // Set up an ItemSampleData object to handle samples and pen channel for each item.
            if (itemsInSubscription != null && itemsInSubscription.Count > 0)
            {
                double offset = 0;
                foreach (var itemInSubscription in itemsInSubscription)
                {
                    ItemSampleData itemData = new ItemSampleData(subscription, scheduler, itemInSubscription.Item2, offset);
                    itemSamplesFromSubscription.Add(itemData);

                    // Adding graph to plotter
                    plotter.AddLineGraph(new CompositeDataSource(xSrc, itemData.ItemSamples),
                        new Pen(ColorPicker.NextColor, 2),
                        new PenDescription(itemInSubscription.Item1 + "(" + itemInSubscription.Item2 + ")"));

                    // Each channel is separated by this much.
                    offset += 250;
                }
            }

            // Turn on the redraw timer.
			timer.Interval = TimeSpan.FromMilliseconds(1000);
            timer.Tick += AnimatedPlot_Timer;
            timer.IsEnabled = true;

            // Force evertyhing plotted to be visible
            plotter.FitToView();
        }

        /// <summary>
        /// Query the OData source, retrieving a list of all items in the
        /// subscription that we monitor.
        /// </summary>
        private void GetItems()
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataLocation);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionNo)?$expand=Items
            var subscriptionWithItems = from subscription in dataContext.Subscriptions.Expand("Items")
                                          where subscription.SubscriptionId == subscriptionNo
                                          select subscription;

            // Execute the query.  There should be exactly one subscription, select only the /Items collection.
            var itemsInSubs = subscriptionWithItems.FirstOrDefault().Items;

            // Capture all the Item Ids in the subscription in the field.
            itemsInSubscription = new List<Tuple<string, int>>();

            foreach (var itemInSubs in itemsInSubs)
            {
                itemsInSubscription.Add(Tuple.Create(itemInSubs.ItemName, itemInSubs.ItemId));
            }
        }

        /// <summary>
        /// When the window is closing, stop our time and clean up.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Closing(object sender, CancelEventArgs e)
        {
            if (timer != null)
            {
                timer.IsEnabled = false;
            }

            itemSamplesFromSubscription = new List<ItemSampleData>();
        }
        #endregion
    }
}
