﻿namespace ODataSampleClient
{
    using System;
    using System.Collections.Generic;
    using System.Data.Services.Client;
    using System.Linq;
    using System.Windows;
    using System.Windows.Controls;

    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window
    {
        #region Private static fields
        /// <summary>
        /// The base endpoint for the service.
        /// </summary>
        private static Uri dataAddress = new Uri("http://localhost:10001/TimeSeriesData.svc");

        /// <summary>
        /// For Create subscription logic.
        /// </summary>
        private static int nextSubscriptionId = 100;
        #endregion

        #region Constructor
        /// <summary>
        /// When the program starts, this constructor starts the main window.
        /// </summary>
        public MainWindow()
        {
            // Required framework initialization.
            InitializeComponent();

            SetIntervalButton.IsEnabled = false;
            DisplaySubscriptionButton.IsEnabled = false;
            DeleteSubscriptionButton.IsEnabled = false;

            // Set a handler into the Loaded event, continue setup after the Window is initialized.
            base.Loaded += Window_Loaded;
        }
        #endregion

        #region Window Loaded event handler
        /// <summary>
        /// Loaded event handler.  When the WPF form is loaded,
        /// this event handler will discover the items in the
        /// subscription, and initialize the pen channels for each item.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            // Update the subscriptions list.
            GetSubscriptions_Button_Click(sender, e);

            // Initial population of the items list is from the service as a whole.
            ItemsFromService.IsChecked = true;
            FillItemsList(0);
        }
        #endregion

        #region Subscription management control code behind
        /// <summary>
        /// Get all subscriptions in the OData service and display them.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void GetSubscriptions_Button_Click(object sender, RoutedEventArgs e)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions
            var subscriptionsQuery = from subscription in dataContext.Subscriptions
                                     select subscription;

            // Execute the query, select subscriptions.
            SubscriptionsList.Items.Clear();
            foreach (var sampleInSubs in subscriptionsQuery)
            {
                // Add each subscription to the list control.
                SubscriptionsList.Items.Add(sampleInSubs.SubscriptionId + "\t" + sampleInSubs.UpdateInterval);
            }
        }

        /// <summary>
        /// When the selected subscription in the list changes, capture the
        /// current one and enable/disable some buttons.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SubscriptionsList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            // If nothing is selected, disable the subscription controls.
            if (SubscriptionsList.SelectedItem == null)
            {
                IntervalToSet.Text = "";
                IntervalToSet.IsEnabled = false;
                SetIntervalButton.IsEnabled = false;
                BrowseButton.IsEnabled = false;
                DisplaySubscriptionButton.IsEnabled = false;
                DeleteSubscriptionButton.IsEnabled = false;

                // Fill the items list from the service.
                ItemsFromSubscription.IsChecked = false;
                ItemsFromService.IsChecked = true;

                ItemsList.Items.Clear();

                // Update the items list with all server items.
                FillItemsList(0);

                return;
            }

            // Capture the selected subscription, and split the id and interval.
            string selectedSubscription = SubscriptionsList.SelectedItem.ToString();
            string[] subsFields = selectedSubscription.Split('\t');

            // If the subscription Id was a valid number, enable the subscription controls.
            int subsId = 0;
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                IntervalToSet.IsEnabled = true;
                IntervalToSet.Text = subsFields[1];
                SetIntervalButton.IsEnabled = true;
                BrowseButton.IsEnabled = true;
                DisplaySubscriptionButton.IsEnabled = true;
                DeleteSubscriptionButton.IsEnabled = true;

                // Fill the items list from this subscription.
                ItemsFromSubscription.IsChecked = true;
                ItemsFromService.IsChecked = false;

                // Update the items list with all subscription items.
                FillItemsList(subsId);
            }
        }

        /// <summary>
        /// The user selects a subscription from the SubscriptionList list
        /// control and presses this button.  Open a new window to display
        /// the items from the selected subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DisplaySubscription_Button_Click(object sender, RoutedEventArgs e)
        {
            if (SubscriptionsList.SelectedItem == null) return;

            string selectedSubscription = SubscriptionsList.SelectedItem.ToString();

            string[] subsFields = selectedSubscription.Split('\t');
            int subsId = 0;
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                Window displayWindow = new DataDisplay(dataAddress, subsId);
                displayWindow.Activate();
                displayWindow.Show();
            }
        }

        /// <summary>
        /// Change the update interval of the selected subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void SetInterval_Button_Click(object sender, RoutedEventArgs e)
        {
            // Capture the selected subscription, and split the id and interval.
            string selectedSubscription = SubscriptionsList.SelectedItem.ToString();
            string[] subsFields = selectedSubscription.Split('\t');

            // If the subscription Id is a valid integer, do the query to change it.
            int subsId = 0;
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                // Always initialize a new context unless you are re-using the query.
                TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

                // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subsId)
                var subscriptionsQuery = from subscription in dataContext.Subscriptions
                                         where subscription.SubscriptionId == subsId
                                         select subscription;

                // There should be exactly one subscription returned from the query.
                TimeSeriesDataFeed.Subscription subs = subscriptionsQuery.FirstOrDefault();

                int newInterval;
                if (subs != null && Int32.TryParse(IntervalToSet.Text, out newInterval))
                {
                    // Make desired changes to the subscription returned from the query.
                    subs.UpdateInterval = newInterval;

                    try
                    {
                        // Tell the context to change a property of the entity it is tracking.
                        dataContext.UpdateObject(subs);

                        // Send changes to the service.
                        dataContext.SaveChanges();

                        // Update the subscriptions list.
                        GetSubscriptions_Button_Click(sender, e);
                    }
                    catch (Exception exception)
                    {
                        Console.WriteLine("Exception {0}", exception.Message);
                    }
                }
            }
        }

        /// <summary>
        /// Delete the selected subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void DeleteSubscription_Button_Click(object sender, RoutedEventArgs e)
        {
            // Capture the selected subscription, and split the id and interval.
            string selectedSubscription = SubscriptionsList.SelectedItem.ToString();
            string[] subsFields = selectedSubscription.Split('\t');

            // If the subscription Id is a valid number, form the query to delete it.
            int subsId = 0;
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                // Always initialize a new context unless you are re-using the query.
                TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

                // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subsId)
                var subscriptionsQuery = from subscription in dataContext.Subscriptions
                                         where subscription.SubscriptionId == subsId
                                         select subscription;

                // There should be exactly one subscription returned from the query.
                TimeSeriesDataFeed.Subscription subs = subscriptionsQuery.FirstOrDefault();

                try
                {
                    // Tell the context to delete one of the entities it is tracking.
                    dataContext.DeleteObject(subs);

                    // Send changes to the service.
                    dataContext.SaveChanges();

                    // Update the subscriptions list.
                    GetSubscriptions_Button_Click(sender, e);
                }
                catch (Exception exception)
                {
                    Console.WriteLine("Exception {0}", exception.Message);
                }
            }
        }

        /// <summary>
        /// Create a new subscription with the next available id.
        /// Refresh the display of subscriptions.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void CreateSubscription_Button_Click(object sender, RoutedEventArgs e)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Create a new client-side subscription object, with desired parameters.
            int subscriptionId = nextSubscriptionId++;
            TimeSeriesDataFeed.Subscription addedSubscription = TimeSeriesDataFeed.Subscription.CreateSubscription(subscriptionId, 0, -1, -1);

            try
            {
                // Add the new object to the context.
                dataContext.AddToSubscriptions(addedSubscription);

                // Send changes to the service.
                DataServiceResponse response = dataContext.SaveChanges();

                // Sample code only, just to explore what the response looks like.
                foreach (ChangeOperationResponse change in response)
                {
                    // Get the descriptor for the entity.
                    EntityDescriptor descriptor = change.Descriptor as EntityDescriptor;

                    if (descriptor != null)
                    {
                        TimeSeriesDataFeed.Subscription newSubscription = descriptor.Entity as TimeSeriesDataFeed.Subscription;

                        if (newSubscription != null)
                        {
                            Console.WriteLine("New subscription added with ID {0}.",
                                newSubscription.SubscriptionId);
                        }
                    }
                }

                ODataSubscription.AddSubscription(subscriptionId);

                // Update the subscriptions list.
                GetSubscriptions_Button_Click(sender, e);
            }
            catch (Exception exception)
            {
                Console.WriteLine("Exception {0}", exception.Message);
            }
        }
        #endregion

        #region Items management controls code behind
        /// <summary>
        /// TODO
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ItemsList_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {

        }

        /// <summary>
        /// When the 'Subscription' radio button is checked,
        /// change the items list to be filled from the current subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ItemsFromSubscription_Checked(object sender, RoutedEventArgs e)
        {
            // Ok, belt and suspenders, but it's cheap...
            if (ItemsFromSubscription.IsChecked != true) return;

            // This will only have an effect if there is really a subscription selected.
            string selectedSubscription = SubscriptionsList.SelectedItem == null ? "" : SubscriptionsList.SelectedItem.ToString();

            // Split the id and interval.
            string[] subsFields = selectedSubscription.Split('\t');

            // If the subscription Id is a valid number, fill the items list from it.
            int subsId = 0;
            string subsInterval = "";
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                subsInterval = subsFields[1];
                FillItemsList(subsId);
            }

            // We can browse this subscription now.
            IntervalToSet.IsEnabled = true;
            IntervalToSet.Text = subsInterval;
            SetIntervalButton.IsEnabled = true;
            BrowseButton.IsEnabled = true;
            DisplaySubscriptionButton.IsEnabled = true;
            DeleteSubscriptionButton.IsEnabled = true;
        }

        /// <summary>
        /// When the 'Service' radio button is checked, change the
        /// items list to be filled from the service.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void ItemsFromService_Checked(object sender, RoutedEventArgs e)
        {
            // Ok, belt and suspenders, but it's cheap...
            if (ItemsFromService.IsChecked != true) return;

            // When ItemsFromService is checked, the subscription Id is ignored, just fill from the service.
            FillItemsList(0);

            // We can not browse and subscription now.
            IntervalToSet.Text = "";
            IntervalToSet.IsEnabled = false;
            SetIntervalButton.IsEnabled = false;
            BrowseButton.IsEnabled = false;
            DisplaySubscriptionButton.IsEnabled = false;
            DeleteSubscriptionButton.IsEnabled = false;

        }

        /// <summary>
        /// Browse the items in the selected subscription, allowing
        /// items from the service to be added or removed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void BrowseButton_Click(object sender, RoutedEventArgs e)
        {
            // This will only have an effect if there is really a subscription selected.
            string selectedSubscription = SubscriptionsList.SelectedItem == null ? "" : SubscriptionsList.SelectedItem.ToString();

            // Split the id and interval.
            string[] subsFields = selectedSubscription.Split('\t');

            // If the subscription Id was a valid number, run the browse dialog.
            int subsId = 0;
            if (subsFields.Length > 1 && Int32.TryParse(subsFields[0], out subsId))
            {
                // Create the browse dialog with the selected subscription, and run as modal.
                Window tagBrowserWindow = new TagBrowser(dataAddress, subsId);
                tagBrowserWindow.ShowDialog();

                // The items in the subscription may have changed, update the list.
                FillItemsList(subsId);
            }
        }
        #endregion

        #region Private helper methods
        /// <summary>
        /// Form the correct query based on the radio buttons,
        /// get the items indicated, and fill the item list box.
        /// </summary>
        /// <param name="subscriptionId"></param>
        private void FillItemsList(int subscriptionId)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            List<TimeSeriesDataFeed.Item> serverItemsList = null;
            if (ItemsFromSubscription.IsChecked == true)
            {
                // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionId)/Items
                var subscriptionQuery = from subscription in dataContext.Subscriptions.Expand("Items")
                                        where subscription.SubscriptionId == subscriptionId
                                        select subscription;

                var serverSubscription = subscriptionQuery.FirstOrDefault();
                if (serverSubscription != null) serverItemsList = serverSubscription.Items.ToList();
            }
            else
            {
                // Generate the URL http://localhost:10001/TimeSeriesData.svc/Items
                var serviceQuery = from item in dataContext.Items
                                   select item;

                serverItemsList = serviceQuery.ToList();
            }

            if (serverItemsList == null) return;

            // Execute the query, select subscriptions.
            ItemsList.Items.Clear();
            foreach (var item in serverItemsList)
            {
                // Add each item to the list control.
                ItemsList.Items.Add(item.ItemName + " (" + item.ItemId + ")");
            }
        }
        #endregion
    }
}
