﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reactive.Concurrency;
using System.Reactive.PlatformServices;
using System.Reactive.Linq;
using System.Reactive.Subjects;
using System.Text;
using System.Timers;
using System.Threading.Tasks;

namespace ODataSampleClient
{
    /// <summary>
    /// Manage a single subscription for multiple consumers.
    /// </summary>
    public class ODataSubscription
    {
        #region Private static fields
        /// <summary>
        /// A single repository for all subscriptions.
        /// This represents the device or backing service, which must persist the subscriptions.
        /// </summary>
        private static Dictionary<int, ODataSubscription> subscriptionRegistry = new Dictionary<int, ODataSubscription>();
        #endregion

        #region Private fields
        /// <summary>
        /// The subscription Id that identifies this subscription
        /// as a resource to the OData service.
        /// </summary>
        private readonly int subscriptionId;

        /// <summary>
        /// Either a provider or a consumer may add a new Subject to
        /// the collection of subjects for this subscription.
        /// A subject implements both IObservable and IObserver.
        /// </summary>
        private Dictionary<int, Subject<List<TimeSeriesDataFeed.Sample>>> sampleFeeds = new Dictionary<int, Subject<List<TimeSeriesDataFeed.Sample>>>();

        /// <summary>
        /// Provide a lock for thread-safe access to sampleFeeds.
        /// </summary>
        private readonly object lockObject = new object();

        /// <summary>
        /// This subscription polls the OData service based on this timer.
        /// </summary>
        private readonly Timer subscriptionPollTimer;
        #endregion

        #region Public static methods
        /// <summary>
        /// Ensure that the static registry contains a subscription with the
        /// specified Id.  If not, create a new one using the supplied interval.
        /// </summary>
        /// <param name="subsId">Ensure that a subscription with this Id exists in the static registry</param>
        /// <param name="interval">If a new subscription needs to be created, use this update interval</param>
        public static void AddSubscription(int subsId, double interval = 1000)
        {
            // If the specified subscription does not exist, create it and add to the registry.
            if (!subscriptionRegistry.ContainsKey(subsId))
            {
                subscriptionRegistry.Add(subsId, new ODataSubscription(subsId, interval));
            }
        }

        /// <summary>
        /// Get a subscription by its id.
        /// </summary>
        /// <param name="subsId">The Id of the subscription to get</param>
        /// <returns>The requested subscription, or null if not found</returns>
        public static ODataSubscription GetSubscription(int subsId)
        {
            // Return it if we've got it.
            ODataSubscription result;
            if (subscriptionRegistry.TryGetValue(subsId, out result)) return result;

            return null;
        }
        #endregion

        #region Private Constructor
        /// <summary>
        /// Initialize a new instance of the <see cref=" ODataSubscription"/> class.
        /// This cannot be called by other classes.  The preferred way to create a
        /// subscription is the public static method AddSubscription.
        /// </summary>
        /// <param name="subsId">The Id of the subscription being created</param>
        /// <param name="interval">The update interval of the subscription</param>
        private ODataSubscription(int subsId, double interval = 1000)
        {
            subscriptionId = subsId;

            subscriptionPollTimer = new Timer(interval);
            subscriptionPollTimer.Elapsed += subsPollHandler;
            subscriptionPollTimer.Enabled = true;
        }
        #endregion

        #region Public Interface
        /// <summary>
        /// Get an IObservable for the specified item Id within this subscription.
        /// The IObservable will be a source of samples for the specified item.
        /// </summary>
        /// <param name="itemId">The item Id identifying the item as a resource within the OData service</param>
        /// <returns>An IObservable interface to a sequence of samples for the specified item</returns>
        public IObservable<List<TimeSeriesDataFeed.Sample>> GetItemSampleFeed(int itemId)
        {
            // A Subject for the specified item Id.  We will return the IObservable interface of this Subject.
            Subject<List<TimeSeriesDataFeed.Sample>> result;

            // Do this in a thread-safe way.
            lock (lockObject)
            {
                // Return it if we already have it.
                if (sampleFeeds.TryGetValue(itemId, out result))
                {
                    return result;
                }

                // Make a new Subject if it did not already exist.
                result = new Subject<List<TimeSeriesDataFeed.Sample>>();
                sampleFeeds.Add(itemId, result);
            }

            return result;
        }

        /// <summary>
        /// Get an IObserver for the specified item Id within this subscription.
        /// The IObserver will accept a sequence of samples for the specified item,
        /// and re-publish to any subscribers it has.
        /// </summary>
        /// <param name="itemId">The item Id identifying the item as a resource within the OData service</param>
        /// <returns>An IObserver interface to a sequence of samples for the specified item</returns>
        public IObserver<List<TimeSeriesDataFeed.Sample>> GetItemObserver(int itemId)
        {
            // A Subject for the specified item Id.  We will return the IObserver interface of this Subject.
            Subject<List<TimeSeriesDataFeed.Sample>> result;

            // Do this in a thread-safe way.
            lock (lockObject)
            {
                // Return it if we already have it.
                if (sampleFeeds.TryGetValue(itemId, out result))
                {
                    return result;
                }

                // Make a new subject if it did not alreaady exist.
                result = new Subject<List<TimeSeriesDataFeed.Sample>>();
                sampleFeeds.Add(itemId, result);
            }

            return result;
        }
        #endregion

        #region Private helper methods
        /// <summary>
        /// Timer handler.  Periodically poll the OData service and process the results.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        private void subsPollHandler(object sender, ElapsedEventArgs args)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(new Uri("http://localhost:10001/TimeSeriesData.svc"));

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionNo)?$expand=Samples
            var subscriptionWithSamplesQuery = (from subscription in dataContext.Subscriptions.Expand("Samples")
                                                where subscription.SubscriptionId == subscriptionId
                                                select subscription).FirstOrDefault();

            // Group the samples by item Id.  This executes the previous query, which calls the OData service.
            var samplesInSubs = from sample in subscriptionWithSamplesQuery.Samples
                                group sample by sample.ItemIdSampled into sampleSubs
                                select sampleSubs;

            // Each group may contain multiple samples for a single item.
            foreach (var subsGroup in samplesInSubs)
            {
                // Get the IObserver for this item Id.  Pass the samples to the obervers as an observable sequence.
                GetItemObserver(subsGroup.Key).OnNext(subsGroup.ToList());
            }
        }
        #endregion
    }
}
