﻿namespace ODataSampleClient
{
    using System;
    using System.Collections.Generic;
    using System.Data.Services.Client;
    using System.Linq;
    using System.Windows;
    using System.Windows.Controls;

    /// <summary>
    /// Interaction logic for TagBrowser.xaml
    /// </summary>
    public partial class TagBrowser : Window
    {
        #region Private fields
        /// <summary>
        /// The base endpoint of the service.
        /// </summary>
        private Uri dataAddress = new Uri("http://localhost:10001/TimeSeriesData.svc");

        /// <summary>
        /// The subscription we browse tags for.
        /// </summary>
        private int subscriptionId = 0;

        /// <summary>
        /// When an item is selected in the service list, capture its Id here.
        /// </summary>
        private int selectedServiceItemId = 0;

        /// <summary>
        /// When an item is selected in the subscription list, capture its id here.
        /// </summary>
        private int selectedSubscriptionItemId = 0;

        /// <summary>
        /// In order to filter out items already in the subscription when
        /// displaying items remaining in the service, put all subscription
        /// item Ids here.
        /// </summary>
        private List<int> subscriptionItemsList = null;
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes a new instance of the <see cref="TagBrowser"/> class.
        /// Typically called from another window and run as modal.
        /// </summary>
        /// <param name="endpoint">The base endpoint address of the service</param>
        /// <param name="subs">The subscription to browse items for</param>
        public TagBrowser(Uri endpoint, int subs)
        {
            // Capture the input parameters.
            dataAddress = endpoint;
            subscriptionId = subs;

            // Required framework initialization.
            InitializeComponent();

            // Update the label above the subscription list box, disable both arrow buttons.
            ItemsInSubsLabel.Content = String.Format("Items in Subscription {0}", subscriptionId);
            RightArrowButton.IsEnabled = false;
            LeftArrowButton.IsEnabled = false;

            // Set a handler into the Loaded event, continue setup after the Window is initialized.
            base.Loaded += Window_Loaded;
        }
        #endregion

        #region Window Loaded event handler
        /// <summary>
        /// Loaded event handler.  When the WPF form is loaded,
        /// this event handler will discover the items in the
        /// subscription, and initialize the pen channels for each item.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            // Always fill subscription items first, so they may be excluded from service items.
            FillSubscriptionItemsList();
            FillServiceItemsList();
        }
        #endregion

        #region Subscription tag browser controls code behind
        /// <summary>
        /// When an item in the service is selected, allow it
        /// to be moved to the subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Items_In_Service_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            // If no item is selected, neither arrow button should work.
            if (ItemsInServiceListbox.SelectedItem == null)
            {
                RightArrowButton.IsEnabled = false;
                LeftArrowButton.IsEnabled = false;
                return;
            }

            // Get the selected service item, parse out the item Id.
            string selectedItem = ItemsInServiceListbox.SelectedItem.ToString();

            int leftParenIndex = selectedItem.IndexOf('(');
            int righParenIndex = selectedItem.IndexOf(')');
            if (leftParenIndex == -1 || righParenIndex == -1) return;
            string itemIdString = selectedItem.Substring(leftParenIndex + 1, righParenIndex - leftParenIndex - 1);

            // If the item Id is a valid number, capture its Id and enable the right-arrow button.
            int itemId = 0;
            if (Int32.TryParse(itemIdString, out itemId))
            {
                selectedServiceItemId = itemId;
                RightArrowButton.IsEnabled = true;
                LeftArrowButton.IsEnabled = false;
            }
        }

        /// <summary>
        /// Only enabled if a service item is selected, move the
        /// selected service item to the subscription.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void RightArrowButton_Click(object sender, RoutedEventArgs e)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Items(selectedServiceItemId)
            var serviceItemQuery = from item in dataContext.Items
                                   where item.ItemId == selectedServiceItemId
                                   select item;

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionId)?$expand="Items"
            var serviceSubsQuery = from subs in dataContext.Subscriptions.Expand("Items")
                                   where subs.SubscriptionId == subscriptionId
                                   select subs;

            // Execute the two queries.  There should be exactly one result from each query.
            TimeSeriesDataFeed.Item selectedItem = serviceItemQuery.FirstOrDefault();
            TimeSeriesDataFeed.Subscription selectedSubscription = serviceSubsQuery.FirstOrDefault();

            if (selectedItem != null && selectedSubscription != null)
            {
                try
                {
                    // Add the new item with a link to the related order.

                    // This is what I think should do it, but does not seem to propagate to the service.
                    //selectedSubscription.Items.Add(selectedItem);

                    // This is the kludge workaround I created to get this to work.
                    selectedSubscription.ItemToAdd = selectedItem.ItemId;

                    dataContext.UpdateObject(selectedSubscription);

                    // Send the changes to the data service.
                    DataServiceResponse response = dataContext.SaveChanges();

                    // Redraw the item lists.
                    FillSubscriptionItemsList();
                    FillServiceItemsList();
                }
                catch (Exception exception)
                {
                    Console.WriteLine("{0}", exception.Message);
                }
            }
        }

        /// <summary>
        /// When an item in the subscription is selected, allow it
        /// to be moved out of the subscription, causing it to be 
        /// shown in the service again.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Items_In_Subscription_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            // If no item is selected, neither arrow button should work.
            if (ItemsInSubscriptionListbox.SelectedItem == null)
            {
                RightArrowButton.IsEnabled = false;
                LeftArrowButton.IsEnabled = false;
                return;
            }

            // Get the selected subscription item, parse out the item Id.
            string selectedItem = ItemsInSubscriptionListbox.SelectedItem.ToString();
            int leftParenIndex = selectedItem.IndexOf('(');
            int righParenIndex = selectedItem.IndexOf(')');
            if (leftParenIndex == -1 || righParenIndex == -1) return;
            string itemIdString = selectedItem.Substring(leftParenIndex + 1, righParenIndex - leftParenIndex - 1);

            // If the item Id is a valid number, capture its Id and enable the left-arrow button.
            int itemId = 0;
            if (Int32.TryParse(itemIdString, out itemId))
            {
                selectedSubscriptionItemId = itemId;
                LeftArrowButton.IsEnabled = true;
                RightArrowButton.IsEnabled = false;
            }
        }

        /// <summary>
        /// Only enabled if a subscription item is selected, remove the
        /// selected subscription item from the subscription, allowing it
        /// to show as a service item.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void LeftArrow_Button_Click(object sender, RoutedEventArgs e)
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionId)?$expand="Items"
            var serviceSubsQuery = from subs in dataContext.Subscriptions.Expand("Items")
                                   where subs.SubscriptionId == subscriptionId
                                   select subs;

            // Execute the query.  There should be exactly one result.
            TimeSeriesDataFeed.Subscription selectedSubscription = serviceSubsQuery.FirstOrDefault();

            if (selectedSubscription != null)
            {
                try
                {
                    // Remove the selected item from the subscription.

                    // This is the kludge workaround I created to get this to work.
                    selectedSubscription.ItemToRemove = selectedSubscriptionItemId;

                    dataContext.UpdateObject(selectedSubscription);

                    // Send the changes to the data service.
                    DataServiceResponse response = dataContext.SaveChanges();

                    // Redraw the item lists.
                    FillSubscriptionItemsList();
                    FillServiceItemsList();
                }
                catch (Exception exception)
                {
                    Console.WriteLine("{0}", exception.Message);
                }
            }
        }
        #endregion

        #region Private helper methods
        /// <summary>
        /// Repaint the subscription item list from a fresh query.
        /// </summary>
        private void FillSubscriptionItemsList()
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Subscriptions(subscriptionId)/Items?$expand="Items"
            var subscriptionQuery = from subscription in dataContext.Subscriptions.Expand("Items")
                                    where subscription.SubscriptionId == subscriptionId
                                    select subscription;

            // There should be exactly one subscription from the query.
            var serverSubscription = subscriptionQuery.FirstOrDefault();

            // Convert the enumeration of Item objects to a list of integer Ids.
            subscriptionItemsList = null;
            if (serverSubscription != null) subscriptionItemsList = (from item in serverSubscription.Items select item.ItemId).ToList();

            if (subscriptionItemsList == null) return;

            // Fill the list with items from the subscription.
            ItemsInSubscriptionListbox.Items.Clear();
            foreach (var item in serverSubscription.Items)
            {
                // Add each item to the list control.
                ItemsInSubscriptionListbox.Items.Add(item.ItemName + " (" + item.ItemId + ")");
            }
        }

        /// <summary>
        /// Repaint the service item list from a fresh query.
        /// </summary>
        private void FillServiceItemsList()
        {
            // Always initialize a new context unless you are re-using the query.
            TimeSeriesDataFeed.TimeSeriesEntities dataContext = new TimeSeriesDataFeed.TimeSeriesEntities(dataAddress);

            // Generate the URL http://localhost:10001/TimeSeriesData.svc/Items
            var serviceQuery = from item in dataContext.Items
                               select item;

            // Execute the query and collect the Item objects in a list.
            List<TimeSeriesDataFeed.Item> serverItemsList = serviceQuery.ToList();
            if (serverItemsList == null) return;

            // If the subscription item Ids have been previously collected, filter them out of this list.
            if (subscriptionItemsList != null)
            {
                // Select only items from the service that are not also in the subscription.
                var serviceItemsExcludingSubscriptionItemsQuery = from item in serverItemsList
                                                                  where !subscriptionItemsList.Contains(item.ItemId)
                                                                  select item;

                // Replace the original Item list with the filtered one.
                serverItemsList = serviceItemsExcludingSubscriptionItemsQuery.ToList();
            }

            // Fill the list with (filtered) items from the subscription.
            ItemsInServiceListbox.Items.Clear();
            foreach (var item in serverItemsList)
            {
                // Add each item to the list control.
                ItemsInServiceListbox.Items.Add(item.ItemName + " (" + item.ItemId + ")");
            }
        }
        #endregion
    }
}
