﻿// Header
//
namespace ODataService
{
    using System;
    using System.Collections.Generic;
    using System.Data.Services;
    using System.Data.Services.Common;
    using System.Linq;

    using ODataService.Simulation;

    #region Time series data provider service
    /// <summary>
    /// Our service data provider identifies the entity model this service exposes 
    /// as the parameter of the DataService base class.
    /// </summary>
    public class TimeSeriesDataService : DataService<TimeSeriesEntities>
    {
        /// <summary>
        /// This method is called only once to initialize service-wide policies.
        /// </summary>
        /// <param name="config">An object that describes the configuration of the service</param>
        public static void InitializeService(DataServiceConfiguration config)
        {
            // Grant only the rights needed to support the client application.
            config.SetEntitySetAccessRule("*", EntitySetRights.AllRead | EntitySetRights.AllWrite);

            config.DataServiceBehavior.MaxProtocolVersion = DataServiceProtocolVersion.V2;
        }
    }
    #endregion

    #region Time series data provider entity model
    /// <summary>
    /// The entity model for this OData service.  At the root, provide a 
    /// collection of items and a collection of subscriptions.  The OData framework 
    /// will examine this class for public properties that return IQueryable of T.
    /// </summary>
    public partial class TimeSeriesEntities
    {
        /// <summary>
        /// Gets the current set of samples from the entity root.  Always empty.
        /// </summary>
        public IQueryable<Sample> Samples
        {
            get { return (new List<Sample>()).AsQueryable<Sample>(); }
        }

        /// <summary>
        /// Gets a queryable collection of items in the device.
        /// </summary>
        public IQueryable<Item> Items
        {
            get
            {
                Console.WriteLine("Returning {0} Items", ItemRegistry.Items.Count);
                return (from item in ItemRegistry.Items 
                        select new Item(item.Id, 
                                        item.Name, 
                                        item.Value, 
                                        item.Time, 
                                        (int)item.Quality))
                       .AsQueryable<Item>();
            }
        }

        /// <summary>
        /// Gets a queryable collection of subscriptions in the device.
        /// </summary>
        public IQueryable<Subscription> Subscriptions
        {
            get
            {
                // TODO - Enhance SubscriiptionRegistry to persist to disk.
                Console.WriteLine("Returning {0} Subscriptions", SubscriptionRegistry.Subscriptions.Count);
                return (from subscription in SubscriptionRegistry.Subscriptions 
                        select new Subscription() { SubscriptionId = subscription.Id })
                        .AsQueryable<Subscription>();
            }
        }
    }
    #endregion

    #region The Sample class exposed to the OData service
    [DataServiceEntity]
    [DataServiceKey("SampleId")]
    /// <summary>
    /// When a collection of samples is exposed, this is the service type.
    /// </summary>
    public class Sample
    {
        /// <summary>
        /// The id used by the client to enumerate a collection.
        /// </summary>
        public int SampleId { get; set; }

        /// <summary>
        /// The item Id this is a sample of.
        /// </summary>
        public int ItemIdSampled { get; set; }

        /// <summary>
        /// The item's current value.
        /// </summary>
        public int SampleValue { get; private set; }

        /// <summary>
        /// The item's current timestamp.
        /// </summary>
        public DateTime SampleTimeStamp { get; private set; }

        /// <summary>
        /// The item's current quality.
        /// TODO: Can we use an enum here?
        /// </summary>
        public int SampleQuality { get; private set; }

        /// <summary>
        /// When converting internal data to this data model, use this constructor.
        /// </summary>
        /// <param name="v">Value</param>
        /// <param name="t">Time stamp</param>
        /// <param name="q">Quality</param>
        public Sample(int id, int itemId, int v, DateTime t, int q)
        {
            SampleId = id;

            ItemIdSampled = itemId;
            SampleValue = v;
            SampleTimeStamp = t;
            SampleQuality = q;
        }
    }
    #endregion

    #region The Item class exposed to the OData service
    /// <summary>
    /// Our OData service exposes a collection of Items.
    /// This is the data model of an item, with the ItemId as the key.
    /// </summary>
    [DataServiceKey("ItemId")]
    public class Item
    {
        private static int nextItemId = 1000;

        /// <summary>
        /// OData queries use the id as a key.
        /// </summary>
        public int ItemId { get; set; }

        /// <summary>
        /// The item's name.
        /// </summary>
        public string ItemName { get; set; }

        /// <summary>
        /// The item's current value.
        /// </summary>
        public int Value { get; set; }

        /// <summary>
        /// The item's current timestamp.
        /// </summary>
        public DateTime TimeStamp { get; private set; }

        /// <summary>
        /// The item's current quality.
        /// TODO: Can we use an enum here?
        /// </summary>
        public int Quality { get; private set; }

        /// <summary>
        /// When creating a new resource of this type, use this constructor.
        /// </summary>
        public Item()
        {
            ItemId = ++nextItemId;
            ItemName = "";
            Value = 0;
            TimeStamp = DateTime.Now;
            Quality = (int)DataQuality.QaulityUncertain;
        }

        /// <summary>
        /// When converting internal data to this data model, use this constructor.
        /// </summary>
        /// <param name="id">The item id of this item</param>
        /// <param name="value">The value for this item</param>
        /// <param name="timestamp">The timestampe for this item</param>
        /// <param name="quality">The quality for this item</param>
        /// <param name="isNew">The new data available flag for this item</param>
        public Item(int id, string name, int value, DateTime timestamp, int quality)
        {
            ItemId = id;
            ItemName = name;
            Value = value;
            TimeStamp = timestamp;
            Quality = quality;
        }
    }
    #endregion

    #region The Subscription class exposed to the OData service
    /// <summary>
    /// Our OData service exposes a collection of Subscriptions.
    /// This is the data model of a subscription, with the SubscriptionId as the key.
    /// </summary>
    [DataServiceKey("SubscriptionId")]
    public class Subscription
    {
        private DataSubscription dataSubscription = null;

        /// <summary>
        /// OData queries use the id as a key.
        /// </summary>
        public int SubscriptionId
        {
            get
            {
                // Refer to the Subscription registry, return the underlying subscription's update interval.
                return (dataSubscription == null) ? 0 : dataSubscription.Id;
            }
            set
            {
                // Refer to the Subscription registry, change the underlying subscription.
                dataSubscription = SubscriptionRegistry.GetSubscription(value);
            }
        }

        /// <summary>
        /// Gets or sets the update interval of this subscription.
        /// Expose a dynamically-computed update interval.
        /// </summary>
        public int UpdateInterval
        {
            get
            {
                // Refer to the Subscription registry, return the underlying subscription's update interval.
                return (dataSubscription == null) ? 0 : dataSubscription.UpdateInterval;
            }
            set
            {
                // Refer to the Subscription registry, change the underlying subscription's update interval.
                if (dataSubscription != null) dataSubscription.UpdateInterval = value;
            }
        }

        /// <summary>
        /// Gets the items that are in this subscription.  
        /// Expose a dynamically-computed item list.
        /// </summary>
        public IQueryable<Item> Items
        {
            get
            {
                // A return value, may be empty but will not be null.
                var result = new List<Item>();

                if (dataSubscription != null)
                {
                    // Refer to the Subscription registry, get the item collection for the subscription.
                    IList<DataItem> items = dataSubscription.Items;
                    foreach (DataItem item in items)
                    {
                        // Convert the item to an OData item model.
                        result.Add(new Item(item.Id, item.Name, item.Value, item.Time, (int)item.Quality));
                    }
                }

                // Sub-queries are supported within the items set.
                Console.WriteLine("Returning {0} Subscription({1})/Items", result.Count, SubscriptionId);
                return result.AsQueryable();
            }
        }

        /// <summary>
        /// Gets the samples accumulated since the last read.
        /// Expose a dynamically-computed samples list of all items in the subscription.
        /// </summary>
        public IQueryable<Sample> Samples
        {
            get
            {
                // A return value, may be empty but will not be null.
                var result = new List<Sample>();

                if (dataSubscription != null)
                {
                    IEnumerable<DataSample> samples = dataSubscription.Samples;
                    foreach (DataSample sample in samples)
                    {
                        result.Add(new Sample(sample.Id, sample.ItemId, sample.Value, sample.Time, (int)sample.Quality));
                    }
                }

                // Sub-queries are supported within the Vtqs set.
                Console.WriteLine("Returning {0} Subscriptions({1})/Samples", result.Count, SubscriptionId);
                return result.AsQueryable();
            }
        }

        /// <summary>
        /// Gets or sets a pseudo-value, which acts like an operation.
        /// Getting always returns 0.  Setting adds the item specified
        /// by Id to the items monitored by this subscription.
        /// </summary>
        public int ItemToAdd
        {
            get
            {
                // Not meaningful, always return -1.
                return -1;
            }
            set
            {
                // If this subscription is active, add a reference to
                // the requested item to our subscribed items.
                if (dataSubscription != null)
                {
                    DataItem itemToAdd = ItemRegistry.GetItem(value);

                    if (itemToAdd != null)
                    {
                        dataSubscription.Items.Add(itemToAdd);
                    }
                }
            }
        }

        /// <summary>
        /// Gets or sets a pseudo-value, which acts like an operation.
        /// Getting always returns 0.  Setting removes the item specified
        /// by Id from the items monitored by this subscription.
        /// </summary>
        public int ItemToRemove
        {
            get
            {
                // Not meaningful, always return -1.
                return -1;
            }
            set
            {
                // If this subscription is active, remove the reference
                // to the requested item from our subscribed items.
                if (dataSubscription != null)
                {
                    foreach (DataItem item in dataSubscription.Items)
                    {
                        if (item.Id == value)
                        {
                            dataSubscription.Items.Remove(item);
                            break;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Discovered through reflection, allows this subscription
        /// to be deleted through a query.
        /// </summary>
        /// <returns>True if successfully removed from the entity model</returns>
        public bool Delete()
        {
            if (dataSubscription != null)
            {
                return SubscriptionRegistry.DeleteSubscription(dataSubscription.Id);
            }

            return false;
        }
    }
    #endregion
}
