﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data.Services;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

using ODataService.Simulation;

namespace ODataService
{
    /// <summary>
    /// The entity model for this OData service.
    /// At the root, provide a collection of items and a collection of subscriptions.
    /// The OData framework will examine this class for public properties, and
    /// create queriable collections if they return IQueryable of T.
    /// </summary>
    public partial class TimeSeriesEntities : IUpdatable
    {
        // Summary:
        //     Adds the specified value to the collection.
        //
        // Parameters:
        //   targetResource:
        //     Target object that defines the property.
        //
        //   propertyName:
        //     The name of the collection property to which the resource should be added..
        //
        //   resourceToBeAdded:
        //     The opaque object representing the resource to be added.
        void IUpdatable.AddReferenceToCollection(object targetResource, string propertyName, object resourceToBeAdded)
        {
            Console.WriteLine("IUpdatable.AddReferenceToCollection({0})", propertyName);

            PropertyInfo requestedProperty = targetResource.GetType().GetProperty(propertyName);
            if (requestedProperty == null)
                throw new Exception(String.Format("Can't find property {0} on resource {1}", propertyName, targetResource.GetType()));

            IList collection = (IList)requestedProperty.GetValue(targetResource, null);
            collection.Add(resourceToBeAdded);
        }

        //
        // Summary:
        //     Cancels a change to the data.
        void IUpdatable.ClearChanges()
        {
            Console.WriteLine("IUpdatable.ClearChanges()");
        }

        //
        // Summary:
        //     Creates the resource of the specified type and that belongs to the specified
        //     container.
        //
        // Parameters:
        //   containerName:
        //     The name of the entity set to which the resource belongs.
        //
        //   fullTypeName:
        //     The full namespace-qualified type name of the resource.
        //
        // Returns:
        //     The object representing a resource of specified type and belonging to the
        //     specified container.
        object IUpdatable.CreateResource(string containerName, string fullTypeName)
        {
            Console.WriteLine("IUpdatable.CreateResource({0}, {1})", containerName, fullTypeName);

            Type t = Type.GetType(fullTypeName, true);
            return Activator.CreateInstance(t);
        }

        //
        // Summary:
        //     Deletes the specified resource.
        //
        // Parameters:
        //   targetResource:
        //     The resource to be deleted.
        void IUpdatable.DeleteResource(object targetResource)
        {
            Console.WriteLine("IUpdatable.DeleteResource({0})", targetResource.GetType().ToString());

            MethodInfo deleteMethod = targetResource.GetType().GetMethod("Delete");
            if (deleteMethod != null && deleteMethod.IsPublic)
            {
                deleteMethod.Invoke(targetResource, null);
            }
        }

        //
        // Summary:
        //     Gets the resource of the specified type identified by a query and type name.
        //
        // Parameters:
        //   query:
        //     Language integrated query (LINQ) pointing to a particular resource.
        //
        //   fullTypeName:
        //     The fully qualified type name of resource.
        //
        // Returns:
        //     An opaque object representing a resource of the specified type, referenced
        //     by the specified query.
        object IUpdatable.GetResource(IQueryable query, string fullTypeName)
        {
            Console.WriteLine("IUpdatable.GetResource({0})", fullTypeName);

            object resource = query.Cast<object>().SingleOrDefault();

            // fullTypeName can be null for deletes 
            if (fullTypeName != null && resource.GetType().FullName != fullTypeName)
                throw new ApplicationException("Unexpected type for this resource.");
            return resource;
        }

        //
        // Summary:
        //     Gets the value of the specified property on the target object.
        //
        // Parameters:
        //   targetResource:
        //     An opaque object that represents a resource.
        //
        //   propertyName:
        //     The name of the property whose value needs to be retrieved.
        //
        // Returns:
        //     The value of the object.
        object IUpdatable.GetValue(object targetResource, string propertyName)
        {
            Console.WriteLine("IUpdatable.GetValue({0}, {1})", targetResource.GetType().ToString(), propertyName);

            // Get the named property for the target resource, if it exists.
            PropertyInfo requestedProperty = targetResource.GetType().GetProperty(propertyName);
            if (requestedProperty != null && requestedProperty.CanRead)
            {
                // The named property exists and is readable, read it.
                return requestedProperty.GetValue(targetResource);
            }

            return null;
        }

        //
        // Summary:
        //     Removes the specified value from the collection.
        //
        // Parameters:
        //   targetResource:
        //     The target object that defines the property.
        //
        //   propertyName:
        //     The name of the property whose value needs to be updated.
        //
        //   resourceToBeRemoved:
        //     The property value that needs to be removed.
        void IUpdatable.RemoveReferenceFromCollection(object targetResource, string propertyName, object resourceToBeRemoved)
        {
            Console.WriteLine("IUpdatable.RemoveReferenceFromCollection({0})", propertyName);
        }

        //
        // Summary:
        //     Resets the resource identified by the parameter resource to its default value.
        //
        // Parameters:
        //   resource:
        //     The resource to be updated.
        //
        // Returns:
        //     The resource with its value reset to the default value.
        object IUpdatable.ResetResource(object resource)
        {
            Console.WriteLine("IUpdatable.ResetResource()");

            return null;
        }

        //
        // Summary:
        //     Returns the instance of the resource represented by the specified resource
        //     object.
        //
        // Parameters:
        //   resource:
        //     The object representing the resource whose instance needs to be retrieved.
        //
        // Returns:
        //     Returns the instance of the resource represented by the specified resource
        //     object.
        object IUpdatable.ResolveResource(object resource)
        {
            Console.WriteLine("IUpdatable.ResolveResource()");

            return resource;
        }

        //
        // Summary:
        //     Saves all the changes that have been made by using the System.Data.Services.IUpdatable
        //     APIs.
        void IUpdatable.SaveChanges()
        {
            Console.WriteLine("IUpdatable.SaveChanges()");

            // Rewrite the configuration file.
            Persistence.GenerateDocument();
        }

        //
        // Summary:
        //     Sets the value of the specified reference property on the target object.
        //
        // Parameters:
        //   targetResource:
        //     The target object that defines the property.
        //
        //   propertyName:
        //     The name of the property whose value needs to be updated.
        //
        //   propertyValue:
        //     The property value to be updated.
        void IUpdatable.SetReference(object targetResource, string propertyName, object propertyValue)
        {
            Console.WriteLine("IUpdatable.SetReference({0})", propertyName);
        }

        //
        // Summary:
        //     Sets the value of the property with the specified name on the target resource
        //     to the specified property value.
        //
        // Parameters:
        //   targetResource:
        //     The target object that defines the property.
        //
        //   propertyName:
        //     The name of the property whose value needs to be updated.
        //
        //   propertyValue:
        //     The property value for update.
        void IUpdatable.SetValue(object targetResource, string propertyName, object propertyValue)
        {
            Console.WriteLine("IUpdatable.SetValue({0}, {1})", targetResource.GetType().ToString(), propertyName);

            // Get the named property for the target resource, if it exists.
            PropertyInfo requestedProperty = targetResource.GetType().GetProperty(propertyName);
            if (requestedProperty != null && requestedProperty.CanWrite)
            {
                // The named property exists and is writable, write to it.
                requestedProperty.SetValue(targetResource, propertyValue);
            }
        }
    }
}
