﻿using System;
using System.Collections.Generic;

namespace ODataService.Simulation
{
    /// <summary>
    /// A quality enum used by the backing data provider.
    /// </summary>
    public enum DataQuality
    {
        QualityGood = 0,
        QualityBad = 1,
        QaulityUncertain = 2,

        QualityUninitialized = 32767
    }

    /// <summary>
    /// The item in the backing data provider.
    /// Wraps a reference to the DataSource using the Item Id
    /// as an index into the raw data.
    /// </summary>
    public class DataItem
    {
        #region Private fields
        /// <summary>
        /// The item id.  Also, the index into
        /// the raw data array in the DataSource.
        /// </summary>
        private int itemId;

        /// <summary>
        /// The item name.
        /// </summary>
        private string itemName;

        /// <summary>
        /// The DataSource that supplies raw data for this item.
        /// </summary>
        private DataSource dataSource;

        /// <summary>
        /// The current Value of this item.
        /// </summary>
        private int value;

        /// <summary>
        /// The current Timestamp of this item.
        /// </summary>
        private DateTime timeStamp;

        /// <summary>
        /// The current Quality of this item.
        /// </summary>
        private DataQuality quality;
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the Item Id of this item.
        /// </summary>
        public int Id
        {
            get
            {
                return itemId;
            }
        }

        /// <summary>
        /// Gets the Item Name of this item.
        /// </summary>
        public string Name
        {
            get
            {
                return itemName;
            }
            set
            {
                itemName = value;
            }
        }

        /// <summary>
        /// Gets or sets the current Value of this item.
        /// On set, also update the raw value in the DataSource.
        /// </summary>
        public int Value
        {
            get
            {
                // Return cached value.
                return value;
            }
            set
            {
                // Update the cache, also, write to the raw data source.
                dataSource.WriteItem(itemId, value);
                this.value = value;
            }
        }

        /// <summary>
        /// Gets or sets the current Timestamp of this item.
        /// </summary>
        public DateTime Time
        {
            get
            {
                return timeStamp;
            }
            set
            {
                timeStamp = value;
            }
        }

        /// <summary>
        /// Gets or sets the current Quality of this item.
        /// </summary>
        public DataQuality Quality
        {
            get
            {
                return quality;
            }
            set
            {
                quality = value;
            }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes a new instance of the DataItem class.
        /// Initialize the item Id and DataSource used for raw data.
        /// </summary>
        /// <param name="id">The Item Id of this item</param>
        /// <param name="source">The DataSource this item uses for raw data</param>
        public DataItem(int id, string name, DataSource source)
        {
            // Capture the item Id and data source.
            itemId = id;
            itemName = name;
            dataSource = source;

            // Initialize this item's VTQ to default.
            value = 0;
            timeStamp = DateTime.Now;
            quality = DataQuality.QualityUninitialized;
        }
        #endregion

        #region Public interface methods
        /// <summary>
        /// Do a demand read from the data source and return the VTQ.
        /// </summary>
        /// <returns>The current sample</returns>
        public DataSample Read()
        {
            // Get the new value from the DataSource raw data.
            int newValue = dataSource.ReadItem(itemId);

            // Capture the VTQ for this update.
            this.value = dataSource.ReadItem(itemId);
            this.timeStamp = DateTime.Now;
            this.quality = DataQuality.QualityGood;

            return new DataSample(this);
        }
        #endregion
    }
}
