﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Timers;

namespace ODataService.Simulation
{
    /// <summary>
    /// The backing data provider.
    /// A simple array of int's with a timer thread
    /// periodically apply one or more algorithms
    /// to keep the data changing.
    /// </summary>
    public class DataSource
    {
        #region Private static fields
        /// <summary>
        /// The largest raw data array allowed.
        /// </summary>
        private static int maxItemIndex = 10000;
        #endregion

        #region Private fields
        /// <summary>
        /// The actual size of the array of raw integers.
        /// </summary>
        private int simulationSize = maxItemIndex;

        /// <summary>
        /// For the sawtooth algorithm, true for up, false for down.
        /// </summary>
        private bool sawtoothGoingUp = true;

        /// <summary>
        /// For the sawtooth algorithm, count ticks before reversing.
        /// </summary>
        private int sawtoothDuration = 0;

        /// <summary>
        /// The tick count limit before reversing the sawtooth algorithm.
        /// </summary>
        private int sawtoothLimit = 100;

        /// <summary>
        /// The raw data.
        /// </summary>
        private int[] simulationItems;

        /// <summary>
        /// The moving-data algorithm timer.
        /// </summary>
        private Timer tickTimer;

        /// <summary>
        /// Thread safety lock object.
        /// </summary>
        private object lockObject;
        #endregion

        #region Public static properties
        /// <summary>
        /// Gets the private max item index used in all DataSource instances.
        /// </summary>
        public static int MaxItemIndex { get { return maxItemIndex; } }
        #endregion

        #region Constructor
        /// <summary>
        /// Initialize a new instance of the DataSource class.
        /// Provide an update interval for the simulation
        /// moving-data algorithms.  A timer running at this
        /// interval will operate on the raw data periodically.
        /// </summary>
        /// <param name="updateInterval">The moving-data algorithm timer interval</param>
        public DataSource(uint updateInterval)
        {
            // Initialize the raw data.
            simulationSize = MaxItemIndex;
            simulationItems = new int[simulationSize];

            lockObject = new object();

            // Initialize and start the timer to the specified period.
            tickTimer = new Timer((double)updateInterval);
            tickTimer.Elapsed += tickTimer_Elapsed;
            tickTimer.Enabled = true;
        }
        #endregion

        #region Public interface methods
        /// <summary>
        /// Read the value of the specified item Id.
        /// For speed, no bounds checks are provided,
        /// so this can throw IndexOutOfRange exceptions.
        /// </summary>
        /// <param name="offset">The offset within the raw data array</param>
        /// <returns></returns>
        public int ReadItem(int offset)
        {
            // Thread safe access into the array.
            lock (lockObject)
            {
                return simulationItems[offset];
            }
        }

        /// <summary>
        /// Write to the specified item id.
        /// For speed, no bounds checks are provided,
        /// so this can throw IndexOutOfRange exceptions.
        /// </summary>
        /// <param name="offset">The offset within the raw data array</param>
        /// <param name="value">The new value to write</param>
        public void WriteItem(int offset, int value)
        {
            // Thread safe access into the array.
            lock (lockObject)
            {
                simulationItems[offset] = value;
            }
        }
        #endregion

        #region Data-change algorithm timer tick handler
        /// <summary>
        /// Handle the moving-data timer tick.
        /// For some items, make a ramp, others use a sawtooth.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void tickTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            // Do all moving data with thread safety.
            lock(lockObject)
            {
                int span = 4;

                // Run several algorithms for different groups of items.
                List<Task> tasks = new List<Task>();
                tasks.Add(Task.Factory.StartNew(() => Algorithm1(0, 1, span)));
                tasks.Add(Task.Factory.StartNew(() => Algorithm1(1, 2, span)));
                tasks.Add(Task.Factory.StartNew(() => Algorithm4(2, 1, span)));
                tasks.Add(Task.Factory.StartNew(() => Algorithm4(3, 2, span)));

                Task.WaitAll(tasks.ToArray());
            }

            // Every sawtoothLimit ticks, reverse the sawtooth direction.
            if (++sawtoothDuration > sawtoothLimit)
            {
                sawtoothDuration = 0;
                sawtoothGoingUp = !sawtoothGoingUp;
            }
        }

        /// <summary>
        /// This algorithm ramps up by the gain specified until
        /// the background sawtooth timer reaches its maximum,
        /// then resets back to the baseline value for this item.
        /// Do this for items starting at start, skipping by span.
        /// </summary>
        /// <param name="start">The first item in the array to execute on</param>
        /// <param name="gain">How steep is the ramp</param>
        /// <param name="span">How many items to skip to the next item</param>
        private void Algorithm1(int start, int gain, int span)
        {
            // Operate on only the items indicated.
            for (int i = start; i < simulationSize; i += span)
            {
                // If the background sawtooth timer has reached its limit, reset.
                if (sawtoothDuration == sawtoothLimit)
                {
                    simulationItems[i] -= sawtoothLimit * gain;
                }
                else
                {
                    simulationItems[i] += gain;
                }
            }
        }

        /// <summary>
        /// This algorithm sawtooths up and then down, depending on the
        /// sign of the goingup flag.
        /// Do this for items starting at start, skipping by span.
        /// </summary>
        /// <param name="start">The first item in the array to execute on</param>
        /// <param name="gain">How steep is the ramp</param>
        /// <param name="span">How many items to skip to the next item</param>
        private void Algorithm4(int start, int gain, int span)
        {
            // Operate on only the items indicated.
            for (int i = start; i < simulationSize; i += span)
            {
                simulationItems[i] = sawtoothGoingUp ? simulationItems[i] + gain : simulationItems[i] - gain;
            }
        }
        #endregion
    }
}
