﻿using System;
using System.Collections.Generic;
using System.Timers;

namespace ODataService.Simulation
{
    /// <summary>
    /// The subscription in the backing data provider.
    /// Allow items to be added and removed, poll all items periodically.
    /// </summary>
    public class DataSubscription
    {
        #region Private fields
        /// <summary>
        /// The subscription Id for this subscription.
        /// </summary>
        private int subscriptionId;

        /// <summary>
        /// All the Items currently subscribed by this subscription.
        /// </summary>
        private List<DataItem> items;

        /// <summary>
        /// Periodically poll all subscribed items using this timer.
        /// </summary>
        private Timer subscriptionTimer;

        /// <summary>
        /// Buffer of recent VTQ samples for this item.
        /// </summary>
        //private List<int> samples;
        private List<DataSample> samples;

        /// <summary>
        /// Thread safety lock.
        /// </summary>
        private object lockObject;
        #endregion

        #region Public properties
        /// <summary>
        /// Gets the subscription Id.
        /// </summary>
        public int Id { get { return subscriptionId; } }

        /// <summary>
        /// Gets the count of subscribed items.
        /// </summary>
        public int ItemCount { get { return items.Count;  } }

        /// <summary>
        /// Gets or sets the update interval used by this
        /// subscription to poll all subscribed items.
        /// </summary>
        public int UpdateInterval 
        {
            get
            {
                // If the timer is not running, the interval is zero.
                if (subscriptionTimer == null || !subscriptionTimer.Enabled) return 0;

                // Return the actual timer interval.
                return (int)subscriptionTimer.Interval;
            }
            set
            {
                // Ignore this if there is no timer.
                if (subscriptionTimer != null)
                {
                    // Set the new subscription timer interval, using 0 to means stopped.
                    if (value > 0)
                    {
                        subscriptionTimer.Interval = value;
                        subscriptionTimer.Enabled = true;
                    }
                    else
                    {
                        subscriptionTimer.Enabled = false;
                    }
                }
            }
        }

        /// <summary>
        /// Gets the internal list of subscribed items.
        /// </summary>
        public IList<DataItem> Items
        {
            get
            {
                return items;
            }
        }

        /// <summary>
        /// Gets the current samples for this item, destructively.
        /// </summary>
        public IEnumerable<DataSample> Samples
        {
            get
            {
                lock (lockObject)
                {
                    // Return the current list of samples, start the list over.
                    List<DataSample> result = samples;
                    samples = new List<DataSample>();
                    return result;
                }

            }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes a new instance of the DataSubscription class.
        /// Initialize the subscription Id and interval.
        /// </summary>
        /// <param name="id">The subscription Id of this subscription</param>
        /// <param name="subscriptionInterval">The initial update interval to poll subscribed items</param>
        public DataSubscription(int id, uint subscriptionInterval)
        {
            // Capture this subscription's Id.
            subscriptionId = id;

            // Initialize internal state to default values.
            items = new List<DataItem>();

            // Start a new sample buffer.
            samples = new List<DataSample>();
            lockObject = new object();

            // Start the subscription timer with the specified interval.
            subscriptionTimer = new Timer((double)subscriptionInterval);
            subscriptionTimer.Elapsed += subscriptionTimer_Elapsed;
            subscriptionTimer.Enabled = true;
        }
        #endregion

        #region Public interface methods
        /// <summary>
        /// Allow the client to subscribe a new item.
        /// Only subscribe any item once.
        /// </summary>
        /// <param name="item">The item to add to this subscription</param>
        public void AddItem(DataItem item)
        {
            if (item == null) return;

            // Thread safety.
            bool mustSave = false;
            lock (lockObject)
            {
                // Only add to this subscription if it is not already subscribed.
                if (!items.Contains(item))
                {
                    items.Add(item);
                    mustSave = true;
                }
            }
        }

        /// <summary>
        /// Allow the client to unsubscribe an item.
        /// </summary>
        /// <param name="item">The item to remove from this subscription if it is present</param>
        public void RemoveItem(DataItem item)
        {
            // Thread safety.
            bool mustSave = false;
            lock (lockObject)
            {
                // Remove this item from subscription if it is subscribed.
                if (items.Contains(item))
                {
                    items.Remove(item);
                    mustSave = true;
                }
            }
        }
        #endregion

        #region Subscription polling timer handler
        /// <summary>
        /// Handle timer elapsed events for the subscription timer.
        /// Poll all items currently subscribed.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void subscriptionTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            // Get a local reference to subscribed items thread safely.
            DataItem[] localItems = null;
            lock (lockObject)
            {
                localItems = items.ToArray();
            }

            // Update all subscribed items.
            List<DataSample> addedSamples = new List<DataSample>();
            foreach (DataItem item in localItems)
            {
                DataSample sample = item.Read();
                sample.Time = DateTime.Now;
                addedSamples.Add(sample);
            }

            lock (lockObject)
            {
                // Add all new samples to the end.
                samples.AddRange(addedSamples);

                // Trim the buffer by removing from the front if needed.
                if (samples.Count > 5000)
                {
                    samples.RemoveRange(0, samples.Count - 5000);
                }
            }
        }
        #endregion
    }
}
