﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace ODataService.Simulation
{
    /// <summary>
    /// A static container of all items in the backing data provider.
    /// </summary>
    public static class ItemRegistry
    {
        #region Private static fields
        /// <summary>
        /// The internal container of items, stored by id.
        /// </summary>
        private static Dictionary<int, DataItem> items = new Dictionary<int, DataItem>();

        /// <summary>
        /// The actual simulation with raw data and algorithms.
        /// </summary>
        private static DataSource simulationSource = new DataSource(200);
        #endregion

        #region Public static properties
        /// <summary>
        /// Public accessor for the internal items.
        /// </summary>
        public static IList<DataItem> Items { get { return items.Values.ToList(); } }
        #endregion

        #region Public static methods
        /// <summary>
        /// Make an item with the specified Id.  Even if it already
        /// exists, force its name to the specified name.
        /// </summary>
        /// <param name="id">The Id of the item to make</param>
        /// <param name="name">The name to associate with the item Id.  If the item Id exists, change its name to this.</param>
        /// <returns>The DataItem made and registered, or found, or null if the item Id was out of range</returns>
        public static DataItem MakeItem(int id, string name)
        {
            if (id < DataSource.MaxItemIndex)
            {
                // Attempt to find an existing item from the provided Id.
                DataItem item;
                if (!items.TryGetValue(id, out item))
                {
                    // We could not find the requested item, make it and add it.
                    item = new DataItem(id, name, simulationSource);
                    items.Add(id, item);
                }
                else
                {
                    // The item Id already existed, change its name.
                    item.Name = name;
                }

                // Return the item we made or found.
                return item;
            }

            // The item Id was out of range.
            return null;
        }

        /// <summary>
        /// Get or make an items.  If the item already existed in the registry, 
        /// get a reference.  If not, make a new one and add it.
        /// </summary>
        /// <param name="id">The item Id to get an item for</param>
        /// <returns>The requested item, or null if the item Id does not exist</returns>
        public static DataItem GetItem(int id)
        {
            // A return value.
            DataItem item = null;

            if (id < DataSource.MaxItemIndex)
            {
                // Attempt to find an existing item from the provided Id.
                items.TryGetValue(id, out item);
            }

            // Return whatever we got, including null.
            return item;
        }
        #endregion
    }
}
