﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace ODataService.Simulation
{
    /// <summary>
    /// Provide persistence to the simulation in an Xml backing store.
    /// The simulation is considered a data source separate from the
    /// OData service, and takes the place of an actual device with its own persistence.
    /// </summary>
    public static class Persistence
    {
        #region Private static fields
        /// <summary>
        /// The default value for the folder to keep the backing Xml file in.
        /// </summary>
        private static string storageFolder = @".\";

        /// <summary>
        /// The default value for the file name to write the backing Xml to.
        /// </summary>
        private static string storageFilename = "SubscriptionConfiguration.xml";

        /// <summary>
        /// The item namespace it broken into tanks and pumps, each with these attributes.
        /// </summary>
        private static string[] itemAttributes =
        {
            "Level",
            "Flow",
            "Temperature",
            "HighAlarmLimit",
            "LowAlarmLimit",
            "SoakTime1",
            "SoakTime2",
            "StatusCode"
        };
        #endregion

        #region Public static properties
        /// <summary>
        /// Gets and sets the folder to keep the backing Xml file in.
        /// </summary>
        public static string StorageFolder { get { return storageFolder; } set { storageFolder = value; } }

        /// <summary>
        /// Gets and sets the file name of the backing Xml.
        /// </summary>
        public static string StorageFilename { get { return storageFilename; } set { storageFilename = value; } }
        #endregion

        #region Static constructor
        /// <summary>
        /// Initializes the backing store if it does not exist.
        /// </summary>
        static Persistence()
        {
            if (!File.Exists(Path.Combine(StorageFolder, StorageFilename)))
            {
                GenerateDocument();
            }
        }
        #endregion

        #region Public static interface methods
        /// <summary>
        /// Called to load the simulation with items and subscriptions.
        /// </summary>
        public static void LoadDocument()
        {
            // Read the Xml into an XDocument.
            var storageDocument = XDocument.Load(Path.Combine(StorageFolder, StorageFilename));

            // Enumerate the items in the device, and add them to the simulation.
            foreach (var itemElement in storageDocument.Root.Element("items").Descendants("item"))
            {
                ItemRegistry.MakeItem(Int32.Parse(itemElement.Attribute("id").Value), itemElement.Attribute("name").Value);
            }

            // Enumerate the subscriptions in the device, and add them
            // to the simulation, with their subscribed items.
            foreach (var subsElement in storageDocument.Root.Element("subscriptions").Descendants("subscription"))
            {
                var subscription = SubscriptionRegistry.GetSubscription(Int32.Parse(subsElement.Attribute("id").Value));

                foreach (var itemElement in subsElement.Descendants("item"))
                {
                    subscription.AddItem(ItemRegistry.GetItem(Int32.Parse(itemElement.Attribute("id").Value)));
                }
            }
        }

        /// <summary>
        /// Called to save the current simulation configuration to backing store.
        /// </summary>
        public static void GenerateDocument()
        {
            // Build the whole configuraiton in an XDocument.
            var storageDocument = new XDocument();

            // Add the outer Xml element.
            var outerElement = new XElement("configuration");
            storageDocument.Add(outerElement);

            // Add the items element to the outer element, and add items to it.
            var itemsElement = new XElement("items");
            outerElement.Add(itemsElement);
            GenerateDocumentItems(itemsElement);

            // Add the subscriptions element to the outer element, and add subscriptions to it.
            var subscriptionsElement = new XElement("subscriptions");
            outerElement.Add(subscriptionsElement);
            GenerateDocumentSubscriptions(subscriptionsElement);

            // Save the XDocument to disk.
            SaveDocument(storageDocument);
        }
        #endregion

        #region Private static helper methods
        /// <summary>
        /// Generate the Xml item elements for the items
        /// in the simulation, and add them to the items
        /// element passed in.  If no items exist in the
        /// simulation, generate a starting set.
        /// </summary>
        /// <param name="itemsElement">The items XElement to add item XElements to</param>
        private static void GenerateDocumentItems(XElement itemsElement)
        {
            // Get the items from the simulation.  Generate a starting set if nothing is there.
            var items = ItemRegistry.Items;
            if (items == null || items.Count == 0)
            {
                items = GenerateInitialItems();
            }

            // We can't proceed if no items or no Xml.
            if (items == null || itemsElement == null) return;

            // Add a new XElement for each item.
            foreach (var item in items)
            {
                itemsElement.Add(new XElement("item",
                    new XAttribute("id", item.Id),
                    new XAttribute("name", item.Name)));
            }
        }

        /// <summary>
        /// If no items exist in the simulation, call this
        /// method to generate a starting set of items.
        /// </summary>
        /// <returns>A list of items, in the same way as they would be listed in the simulation</returns>
        private static List<DataItem> GenerateInitialItems()
        {
            // A return list.
            var items = new List<DataItem>();

            int nextItemId = 0;

            // Make some tank items.
            for (int tank = 0; tank < 10; tank++)
            {
                for (int attr = 0; attr < itemAttributes.Length; attr++)
                {
                    // Make an item, including its name.
                    items.Add(new DataItem(nextItemId++, "Tank" + (tank + 1).ToString() + "." + itemAttributes[attr], null));
                }
            }

            // Make some pump items.
            for (int pump = 0; pump < 10; pump++)
            {
                for (int attr = 0; attr < itemAttributes.Length; attr++)
                {
                    // Make an item, including its name.
                    items.Add(new DataItem(nextItemId++, "Pump" + (pump + 1).ToString() + "." + itemAttributes[attr], null));
                }
            }

            return items;
        }

        /// <summary>
        /// Enumerate the subscriptions in the simulation and add
        /// an XElement for each subscription to the subscriptions
        /// XElement passed in.
        /// </summary>
        /// <param name="subscriptionsElement">The subscriptions XElement to add subscription XElements to</param>
        private static void GenerateDocumentSubscriptions(XElement subscriptionsElement)
        {
            // Get a collection of subscriptions currently in the simulation.
            var subscriptions = SubscriptionRegistry.Subscriptions;

            // If no subscriptions or no XElement, we cannot continue.
            if (subscriptions == null || subscriptionsElement == null) return;

            foreach (var subscription in subscriptions)
            {
                // Add the subscription XElement to the subscriptions XElement.
                var subscriptionElement = new XElement("subscription",
                    new XAttribute("id", subscription.Id),
                    new XAttribute("updateInterval", subscription.UpdateInterval));
                subscriptionsElement.Add(subscriptionElement);

                // Each subscriptions XElement has an items XElement child, even if it is empty.
                var itemsElement = new XElement("items");
                subscriptionElement.Add(itemsElement);

                // Add an item XElement to the items XElement for each item in the subscription.
                foreach (var item in subscription.Items)
                {
                    itemsElement.Add(new XElement("item",
                        new XAttribute("id", item.Id),
                        new XAttribute("name", item.Name)));
                }
            }
        }

        /// <summary>
        /// When the XDocument is prepared, write it by calling this method.
        /// </summary>
        /// <param name="storageDocument">The XDocument to write to disk</param>
        private static void SaveDocument(XDocument storageDocument)
        {
            // Ensure the directory exists.
            Directory.CreateDirectory(StorageFolder);

            // Save the document to disk.
            if (storageDocument != null)
            {
                storageDocument.Save(Path.Combine(StorageFolder, StorageFilename));
            }
        }
        #endregion
    }
}
