﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ODataService.Simulation
{
    public static class SampleRegistry
    {
        #region Private fields
        private static int sampleCount = 10000;

        private static DataSample[] samples = null;

        private static HashSet<int> freeSamples = new HashSet<int>();
        private static HashSet<int> assignedSamples = new HashSet<int>();
        private static object lockObject = null;
        #endregion

        #region Public properties
        public static List<DataSample> Samples
        {
            get
            {
                return samples.ToList();
            }
        }
        #endregion

        #region Constructor
        /// <summary>
        /// Static initializer for the static fields.
        /// </summary>
        static SampleRegistry()
        {
            samples = new DataSample[sampleCount];
            freeSamples = new HashSet<int>();
            assignedSamples = new HashSet<int>();
            lockObject = new object();

            for (int i = 0; i < sampleCount; i++)
            {
                samples[i] = new DataSample(i);
                freeSamples.Add(i);
            }
        }

        /// <summary>
        /// Get a free sample for use in sampling an item.
        /// </summary>
        /// <returns>The next free sample, or null if no free samples</returns>
        public static DataSample GetSample()
        {
            // Retain this value if no free sample found.
            int freeId = -1;

            lock (lockObject)
            {
                // If there are any free samples, return one.
                if (freeSamples.Count > 0)
                {
                    // Move the next free Id from free to assigned.
                    freeId = freeSamples.First();
                    freeSamples.Remove(freeId);
                    assignedSamples.Add(freeId);
                }
            }

            // Return what we found or null if nothing found.
            return (freeId == -1) ? null : samples[freeId];
        }

        /// <summary>
        /// Return a sample to freedom.
        /// </summary>
        public static void FreeSample(DataSample sample)
        {
            // Retain this value if no free sample found.
            int freeId = sample.Id;

            lock (lockObject)
            {
                // If there are any free samples, return one.
                if (freeSamples.Count > 0)
                {
                    // Move the next Id from assigned to free.
                    assignedSamples.Remove(freeId);
                    freeSamples.Add(freeId);
                }
            }
        }

        public static DataSample GetAssignedSample(int id)
        {
            lock (lockObject)
            {
                if (assignedSamples.Contains(id))
                {
                    return samples[id];
                }
            }

            return null;
        }
        #endregion

    }
}
