﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace ODataService.Simulation
{
    /// <summary>
    /// A static container of all subscriptions in the backing data provider.
    /// </summary>
    public static class SubscriptionRegistry
    {
        #region Private static fields
        /// <summary>
        /// The internal container of subscriptions, stored by id.
        /// </summary>
        private static Dictionary<int, DataSubscription> subscriptions = new Dictionary<int, DataSubscription>();

        #endregion

        #region Public static properties
        /// <summary>
        /// Public accessor for the internal subscriptions.
        /// </summary>
        public static IList<DataSubscription> Subscriptions { get { return subscriptions.Values.ToList(); } }
        #endregion

        #region Public static methods
        /// <summary>
        /// Get or make a subscription.  If the subscription already existed
        /// in the registry, get a reference.  If not, make a new one and add it.
        /// </summary>
        /// <param name="id">The subscription Id to get a subscription for</param>
        /// <returns>Alway return a subscription, never fails</returns>
        public static DataSubscription GetSubscription(int id)
        {
            // Attempt to find an existing subscription from the provided Id.
            DataSubscription subscription;
            if (!subscriptions.TryGetValue(id, out subscription))
            {
                // We could not find the requested subscription, make it and add it.
                subscription = new DataSubscription(id, 1000);
                subscriptions.Add(id, subscription);
            }

            // Return the subscription we made or found.
            return subscription;
        }

        /// <summary>
        /// Delete the specified subscription, if it exists.
        /// </summary>
        /// <param name="id">The subscription Id to delete</param>
        /// <returns>True if the subscription existed and was deleted, false if it did not exist</returns>
        public static bool DeleteSubscription(int id)
        {
            // Attempt to find an existing subscription from the provided Id.
            DataSubscription subscription;
            if (subscriptions.TryGetValue(id, out subscription))
            {
                // Turn the timer off.
                subscription.UpdateInterval = 0;

                // Remove references to the subscription.
                subscriptions.Remove(id);
            }

            // The requested subscription did not exist.
            return false;
        }
        #endregion

    }
}
